#if defined(WIN32)
#pragma warning( disable : 4786 )
#endif

#include <cmath>
#include "SSCatmullClark.h"
#include "SSQModel.h"

SSCatmullClark::SSCatmullClark()
{
}

SSCatmullClark::~SSCatmullClark()
{
}

SSLath *SSCatmullClark::Subdivision(vector<SSLath*> &vec, SSLath *start)
{
  SSCatmullClark::Subdivide(vec, start);
  if(vec.size()>0)
    return vec[0];
  return 0;
}

void SSCatmullClark::Subdivide(vector<SSLath*> &vec, SSLath *start)
{
  if(start == 0 || start->isvisited())
    return;

  static vector<SSVertex> vs, nvs(4, SSVertex()), fvs, evs;
  static SSVertex tmp, ny;
  static SSLath *nf=0, *tmp2=0, *tmp3=0;
  SSLath *tmp1;
  static int i, c=0, k, in;

  //New Face Vertices
  tmp1 = start;
  do
    {
      tmp1->visited(true);
      vs.clear();
      SSQModel::Instance()->FaceVertices(vs, tmp1);
      k = vs.size();
      ny = SSVertex();
      for(i=0; i<k; i++) 
        {
          tmp = vs[i]/(CoordType)k;
          ny = ny + tmp; 
        }
      fvs.push_back(ny);
      if(!tmp1->ccwv())
        {
          fvs.clear();
          do
            {
              tmp1->visited(true);
              SSCatmullClark::Subdivide(vec, tmp1->ccwf());
              tmp1 = tmp1->cwv();
            }
          while(tmp1);
                
          return;
        }
      tmp1 = tmp1->ccwv();
    }
  while(tmp1 != start);

  //New Edge Vertices
  tmp1 = start;
  k = fvs.size();
  c = 0;
  do
    {
      ny = SSVertex();
      tmp = fvs[c]/4.0;
      ny = ny + tmp;
      tmp = fvs[(c+1) % k]/4.0;
      ny = ny + tmp;
      tmp = tmp1->corner()/4.0;
      ny = ny + tmp;
            
      tmp = tmp1->cwf()->corner()/4.0;
      ny = ny + tmp;
            
      evs.push_back(ny);
      tmp1 = tmp1->ccwv();
      c++;
    }
  while(tmp1 != start);
              
  //New Center Vertex
  ny = start->corner()*(CoordType)((CoordType)(k-2)/(CoordType)k);
    
  tmp1 = start;
  tmp = SSVertex();
  do
    {
      tmp = tmp + tmp1->ccwf()->corner();
      tmp1 = tmp1->ccwv();
    }
  while(tmp1 != start);
        
  ny = ny + tmp/(CoordType)(k*k);
  tmp = SSVertex();
       
  k = fvs.size();
  for(i = 0; i < k; i++)
    tmp = tmp + fvs[i];
        
  ny = ny + tmp/(CoordType)(k*k);
       
  //Build the faces adjacent to the Center vertex
  nvs[0] = ny;
  tmp1 = start;
  for(i = 0; i < k; i++)
    {
      nvs[1] = evs[(k-1+i) % k];
      nvs[2] = fvs[i];
      nvs[3] = evs[i];
            
      nf = SSQModel::Instance()->BuildFace(nvs, vec);
                        
      tmp1->nextlevel(nf->ccwf()->ccwf());
      tmp1 = tmp1->ccwv();
    }
        
  //Connect the faces adjacent to the Center vertex
  tmp1 = start;
  do
    {
      SSQModel::Instance()->ConnectFaces(tmp1->nl(), 2, tmp1->ccwv()->nl(), 3);
      tmp1 = tmp1->ccwv();
    }
  while(tmp1 != start);

  // Start subdivision chain reaction
  tmp1 = start;
  fvs.clear();
  evs.clear();
  do
    {
      SSCatmullClark::Subdivide(vec, tmp1->ccwf());
      tmp1 = tmp1->ccwv();
    }
  while(tmp1 != start);
        
  //Connect new faces
  tmp1 = start;
  do
    {
      if(tmp1->ccwf()->nl())
        {
          SSQModel::Instance()->ConnectFaces(tmp1->nl(), 0, tmp1->ccwf()->nl(), 1);
        } 
      if(tmp1->cwf()->nl())
        {
          SSQModel::Instance()->ConnectFaces(tmp1->nl(), 1, tmp1->cwf()->nl(), 0);
        }
      tmp1 = tmp1->ccwv();
    }
  while(tmp1 != start);      
}
