#if defined(WIN32)
#pragma warning( disable : 4786 )
#endif

#include "SSCornerLath.h"
#include "SSVertex.h"

// See Sederberg et al for labelling of knot spacings
// Request for non-existant knot spacing returns -1
KnotSpacingType SSCornerLath::knotspc(int s, int e, int r) const
{
  SSLath *tmp = const_cast<SSCornerLath*>(this);
  int c;

  // Find correct start vertex
  if(s>=0)
    {
      for(c=s; c>0; c--) { tmp = tmp->cwf(); }

      // Rotate
      if(r>0)
        for(c=r; c>0; c--) {
          tmp = tmp->ccwv();
          if(!tmp) {
            cerr << "1 Returned -1.0    s " << s << "   e " << e 
                 << "   r " << r << endl;
            return (KnotSpacingType) -1.0;
          }
        }
      else
        for(c=r; c<0; c++) {
          tmp = tmp->cwv();
          if(!tmp) {
            cerr << "2 Returned -1.0    s " << s << "   e " << e 
                 << "   r " << r << endl;
            return (KnotSpacingType) -1.0;
          }
        }
      if(e<s) {
        return tmp->ccwknotspc();
      } else if(e>s) {
        return tmp->cwknotspc();
      } else {
        cerr << "3 Returned -1.0    s " << s << "   e " << e 
             << "   r " << r << endl;
        return (KnotSpacingType) -1.0;
      }
    }
  else // s<0
    {
      for(c=s; c<0; c++) { tmp = tmp->ccwf(); }

      // Rotate
      if(r>0)
        for(c=r; c>0; c--) {
          tmp = tmp->ccwv();
          if(!tmp) {
            cerr << "4 Returned -1.0    s " << s << "   e " << e
                 << "   r " << r << endl;
            return (KnotSpacingType) -1.0;
          }
        }
      else
        for(c=r; c<0; c++) {
          tmp = tmp->cwv();
          if(!tmp) {
            cerr << "5 Returned -1.0    s " << s << "   e " << e
                 << "   r " << r << endl;
            return (KnotSpacingType) -1.0;
          }
        }
      if(e<s) {
        return tmp->ccwknotspc();
      } else if(e>s) {
        return tmp->cwknotspc();
      } else {
        cerr << "6 Returned -1.0    s " << s << "   e " << e
             << "   r " << r << endl;
        return (KnotSpacingType) -1.0;
      }
    }
}
        
SSCornerLath::SSCornerLath(SSVertex corner)
{
  m_corner = corner;
  m_ccwnextv = m_ccwnextf = m_nextlevel = 0;
  m_visited = false;
  m_ccwknot = m_cwknot = 1.0;
}


SSCornerLath::~SSCornerLath()
{
}


ostream& SSCornerLath::put(ostream &s) const
{
  s << "V: " << m_corner << " ";
        
  s << "CCWV ";
  if(m_ccwnextv != 0)
    s << m_ccwnextv->corner() << " ";
  else
    s << "null ";

  s << "CCWF ";
  if(m_ccwnextf != 0)
    s << m_ccwnextf->corner() << " ";
  else
    s << "null ";

  s << "NL ";
  if(m_nextlevel != 0)
    s << m_nextlevel->corner();
  else
    s << "null";

  return s;
}
        
ostream& operator<<(ostream &s, SSLath &l)
{
  return l.put(s);
}

ostream& operator<<(ostream &s, SSLath *l)
{
  return l->put(s);
}
