#if defined(WIN32)
#pragma warning( disable : 4786 )
#endif

#include <cmath>
#include "SSCubicNURSS.h"
#include "SSQModel.h"
#include "SSCornerLath.h"

SSCubicNURSS::SSCubicNURSS()
{
}

SSCubicNURSS::~SSCubicNURSS()
{
}

SSLath *SSCubicNURSS::Subdivision(vector<SSLath*> &vec, SSLath *start)
{
  Subdivide(vec, start);
  if(vec.size()>0)
    return vec[0];
  return 0;
}

void SSCubicNURSS::Subdivide(vector<SSLath*> &vec, SSLath *start)
{
  if(start == 0 || start->isvisited())
    return;

  vector<SSVertex> vs, evs, nvs(4), fvs, M;
  vector<SSLath *> els(4, (SSLath *) 0);
  vector<CoefficientType> coeff;
  SSVertex tmp, ny, Mt;
  SSLath *nf=0, *tmp1=0, *tmp2=0;
  int i, c=0, k;
  vector<KnotSpacingType> w, knots, nknots(8);
  KnotSpacingType wi, ws=0, aij, aji, ad, ci, cj, mi, fi;
        
  //New Face Vertices
  tmp1 = start;
  do
    {
      ws=0;
      tmp1->visited(true);
      vs.clear();
      SSQModel::Instance()->FaceVertices(vs, tmp1);
      k = vs.size();
      ny = SSVertex();

      w.clear();
      // Calculate coefficients w_i, and sum of w_i* P_i
      tmp2 = tmp1;
      tmp = SSVertex();
      for(i=0; i<k; i++)
        {
          // See formula (12)+(11) in Sederberg et al.
          wi = (tmp2->knotspc(1,0,0) +
                tmp2->knotspc(1,0,2) +
                tmp2->knotspc(1,0,-2) +
                tmp2->knotspc(-2,-1,0) +
                tmp2->knotspc(-2,-1,2) +
                tmp2->knotspc(-2,-1,-2)) *
            (tmp2->knotspc(-1,0,0) +
             tmp2->knotspc(-1,0,2) +
             tmp2->knotspc(-1,0,-2) +
             tmp2->knotspc(2,1,0) +
             tmp2->knotspc(2,1,2) +
             tmp2->knotspc(2,1,-2));
          w.push_back(wi);
          ws += wi;
          tmp += wi * vs[i];
          tmp2 = tmp2->ccwf();
        }
            
      if(ws < 1e-10)
        {
          tmp = SSVertex();
          for(i=0; i<k; i++)
            {
              tmp += vs[i];
            }
          tmp /= (CoefficientType) k;
          fvs.push_back(tmp);
        }
      else
        fvs.push_back(tmp/ws);

      // If ccwv does not exist, we are on the boundary and there is nothing to do,
      // except the recursive calls on the neighbouring vertices.
      if(!tmp1->ccwv())
        {
          do
            {
              tmp1->visited(true);
              Subdivide(vec, tmp1->ccwf());
              tmp1 = tmp1->cwv();
            }
          while(tmp1);

          return;
        }
      tmp1 = tmp1->ccwv();
    }
  while(tmp1 != start);

  //New Edge Vertices
  tmp1 = start;
  k = fvs.size();
  c = 0;
  do
    {
      // Compute coefficients, see Sederberg et al. eqns. (14-15)
      ad = 2.0 * (tmp1->knotspc(1,0,1) +
                  tmp1->knotspc(0,1,-1) +
                  tmp1->knotspc(1,0,-1) +
                  tmp1->knotspc(0,1,1));

      if(ad < 1e-10) {
        aij = aji = 0.0;
      } else {
        aij = (tmp1->knotspc(1,0,1) + tmp1->knotspc(0,1,-1)) / ad;
        aji = (tmp1->knotspc(0,1,1) + tmp1->knotspc(1,0,-1)) / ad;
      }

      ci = tmp1->knotspc(1,0,0) + tmp1->knotspc(1,0,2) + tmp1->knotspc(1,0,-2);
      cj = tmp1->knotspc(0,1,0) + tmp1->knotspc(0,1,2) + tmp1->knotspc(0,1,-2);
            
      if(ci+cj < 1e-10)
        Mt = (tmp1->corner() + tmp1->cwf()->corner())*0.5;
      else
        Mt = (tmp1->corner() * ci + tmp1->cwf()->corner() * cj) / (ci+cj);

      M.push_back(Mt);
            
      // Sederberg et al. eq (13)
      ny = Mt * (1.0-aij-aji) + fvs[(c+1)%k] * aij + fvs[c] * aji;
                
      evs.push_back(ny);
      tmp1 = tmp1->ccwv();
      c++;
    }
  while(tmp1 != start);
              
  //New Center Vertex
  tmp = SSVertex();
  tmp1 = start;
  ad = 0;
  i = 0;
  do
    {
      // Sederberg et al. eqs (16-19)
      mi = (tmp1->knotspc(0,1,1) + tmp1->knotspc(0,1,-1)) *
        (tmp1->knotspc(0,1,2) + tmp1->knotspc(0,1,-2)) * 0.5;

      fi = tmp1->knotspc(0,1,1) * tmp1->knotspc(0,-1,-1);

      ad += mi+fi;
      tmp += mi * M[i] + fi * fvs[i];
      tmp1 = tmp1->ccwv();
      i++;
    }
  while(tmp1 != start);

  if(ad < 1e-10)
    ny = start->corner();
  else
    ny = ((i-3.0)/i) * start->corner() + (3.0 * tmp) / (((double) i) * ad);

  //Build the faces adjacent to the Center vertex
  nvs[0] = ny;
  tmp1 = start;
  for(i = 0; i < k; i++)
    {
      nvs[1] = evs[(k-1+i) % k];
      nvs[2] = fvs[i];
      nvs[3] = evs[i];

      nknots[0] = 0.5 * tmp1->knotspc(0,1,0);
      nknots[1] = 0.5 * tmp1->knotspc(0,-1,0);
      nknots[2] = nknots[1];
      nknots[3] = 0.25 * (tmp1->knotspc(0,1,0) + tmp1->knotspc(-1,-2,0));
      nknots[4] = nknots[3];
      nknots[5] = 0.25 * (tmp1->knotspc(0,-1,0) + tmp1->knotspc(1,2,0));
      nknots[6] = nknots[5];
      nknots[7] = nknots[0];
            
      nf = SSQModel::Instance()->BuildFace(nvs, nknots, vec);
            
      tmp1->nextlevel(nf->ccwf()->ccwf());
      tmp1 = tmp1->ccwv();
    }
        
  //Connect the faces adjacent to the Center vertex
  tmp1 = start;
  do
    {
      SSQModel::Instance()->ConnectFaces(tmp1->nl(), 2, tmp1->ccwv()->nl(), 3);
      tmp1 = tmp1->ccwv();
    }
  while(tmp1 != start);

  // Start subdivision chain reaction
  tmp1 = start;
  do
    {
      Subdivide(vec, tmp1->ccwf());
      tmp1 = tmp1->ccwv();
    }
  while(tmp1 != start);
        

  //Connect new faces
  tmp1 = start;
  do
    {
      if(tmp1->ccwf()->nl())
        {
          SSQModel::Instance()->ConnectFaces(tmp1->nl(), 0, tmp1->ccwf()->nl(), 1);
        } 
      if(tmp1->cwf()->nl())
        {
          SSQModel::Instance()->ConnectFaces(tmp1->nl(), 1, tmp1->cwf()->nl(), 0);
        }
      tmp1 = tmp1->ccwv();
    }
  while(tmp1 != start);      
}
