#if defined(WIN32)
#pragma warning( disable : 4786 )
#endif

#include <cmath>
#include "SSDooSabin.h"
#include "SSQModel.h"

const CoefficientType SSDooSabin::regcoeff[] = { 0.5625, 0.1875, 0.0625, 0.1875 };

SSDooSabin::SSDooSabin()
{
}

SSDooSabin::~SSDooSabin()
{
}

SSLath *SSDooSabin::Subdivision(vector<SSLath*> &vec, SSLath *start)
{

  SSDooSabin::Subdivide(vec, start);
  return start->nl();
}

vector<CoefficientType> SSDooSabin::CalcCoeff(int valence)
{       
  vector<CoefficientType> coeff;

  if(valence == 4)
    {
      for(int j = 0; j < valence; j++)
        coeff.push_back(regcoeff[j]);

      return coeff;
    }

  coeff.push_back((CoefficientType) (0.25 + 1.25/(CoefficientType)valence));

  for(int i = 1; i < valence; i++)
    coeff.push_back((CoefficientType) ((3.0 + 2.0*cos(2.0*((CoefficientType)i)*c_PI
                    /((CoefficientType) valence)))/((CoefficientType) 4.0*valence)));

  return coeff;
}

void SSDooSabin::Subdivide(vector<SSLath*> &vec, SSLath *start)
{
  // If null or already subdivided just return
  if(start == 0 || start->nl())
    return;

  vector<SSVertex> vs, evs(4, SSVertex()), nvs;
  vector<SSLath *> els(4, (SSLath *) 0);
  vector<CoefficientType> coeff;

  SSLath *nf=0, *tmp1=0, *tmp2=0, *tmp3=0;

  int i, j, c, k;

  SSQModel::Instance()->FaceVertices(vs, start);
  k = vs.size();
  coeff = CalcCoeff(k);

  for(i=0; i<k; i++) 
    {
      SSVertex ny;

      for(j=0; j<k; j++) 
        {
          c = (i+j)%k;
          SSVertex tmp = vs[c]*coeff[j];
          ny = ny + tmp;
        }
      nvs.push_back(ny);
    }

  nf = SSQModel::Instance()->BuildFace(nvs, vec);

  // Connect to subdivided level
  start->nextlevel(nf);

  tmp1 = start;
  tmp2 = nf;
  do 
    {
      tmp1->nextlevel(tmp2);
      tmp1 = tmp1->ccwf();
      tmp2 = tmp2->ccwf();
    } 
  while(tmp1 != start);

  // Start subdivision chain reaction
  tmp3 = start;
  do 
    {
      SSDooSabin::Subdivide(vec, tmp3->ccwv());
      tmp3 = tmp3->ccwf();
    } 
  while(start != tmp3);

  // Build and connect faces corresponding to old edges
  tmp1 = start;
  for(i = 0; i < k; i++, tmp1 = tmp1->ccwf()) 
    {
      if(tmp1->ccwv() && !tmp1->nl()->ccwv()) 
        {
          tmp2 = tmp1;
          evs[3] = tmp2->nl()->corner();
          els[0] = tmp2->nl();
          tmp2 = tmp2->ccwv();
          evs[2] = tmp2->nl()->corner();
          tmp2 = tmp2->ccwf();
          evs[1] = tmp2->nl()->corner();
          els[1] = tmp2->nl();
          tmp2 = tmp2->ccwv();
          evs[0] = tmp2->nl()->corner();

          tmp2 = SSQModel::Instance()->BuildFace(evs, vec);

          SSQModel::Instance()->ConnectFaces(els[0], 0, tmp2, 0);
          SSQModel::Instance()->ConnectFaces(els[1], 0, tmp2, 2);
        }
    }

  // Build and connect faces corresponding to old vertices
  tmp1 = start;

  do 
    {
      if(tmp1->ccwv() && !tmp1->nl()->ccwv()->ccwv())
        { // If not already connected, build & connect face
          tmp2 = tmp1;
          vs.clear();
          do 
            {
              tmp3 = tmp2->nl();
              if(!tmp3 || !tmp3->ccwv()) 
                {
                  /* All faces sharing edges with f have been subdivided in the
                   * recursion, but faces sharing only vertices with f may not
                   * have been.  Solution: Just skip affected vertex -- it will
                   * be processed further down the recursion.
                   **/
                  goto loop;
                }
              else
                vs.push_back(tmp3->corner());

              tmp2 = tmp2->ccwv();

              if(!tmp2)
                goto loop;
            } 
          while(tmp1  != tmp2);

          tmp3 = SSQModel::Instance()->BuildFace(vs, vec);

          do 
            {
              if(tmp2->nl()->ccwv())
                SSQModel::Instance()->ConnectFaces(tmp2->nl()->ccwv(), 0, tmp3, 1);

              tmp2 = tmp2->ccwv();
              tmp3 = tmp3->ccwf();
            } 
          while(tmp1 != tmp2);
        }
    loop:
      tmp1 = tmp1->ccwf();
    }
  while(tmp1 != start);
}
