#if defined(WIN32)
#pragma warning( disable : 4786 )
#endif

#include <iostream>
#include <strstream>
#include <cstdio>
#include "SSHeroReader.h"
#include "SSQModel.h"
#include "SSTypes.h"

const string c_VertexSection = "Coordinate";
const string c_FaceSection = "coordIndex";
const string c_Stop = "Stop";
const int c_HashSize = 1000;
double maxx, minx, maxy, miny, maxz, minz;
int c;

#define MAX(a,b) (a>b ? a : b)
#define MIN(a,b) (a<b ? a : b)

SSHeroReader::SSHeroReader()
{
  m_numberoffaces = 0;
}

SSHeroReader::~SSHeroReader()
{
}


bool SSHeroReader::read(const string filename, vector<SSLath*> &faces)
{
  cout << "Opening file: " << filename << endl;
    
  ifstream file(filename.c_str());

  if(!file.is_open())
    return false;
   
  m_nextsection = c_VertexSection;
  m_lineno = 0;

  while(true)
    {
      switch(ReadLine(file))
        {
        case IO_EOF:
          DumpAll();
          return BuildModel(faces);
        case IO_FAIL:
          cout << "IO Fail" << endl;
          return false;
        case IO_GOOD:
          break;
        case IO_NEXTSECTION:
          cout << "New Section " << m_nextsection << endl;
          DoNextSection(file);
          break;
        default:
          return false;
        }
    }            

  return false;
}

SSHeroReader::io_status SSHeroReader::ReadLine(ifstream &is)
{
  is.getline(m_currentline, 1024);

  if(is.eof() || m_nextsection == c_Stop) 
    return IO_EOF;
    
  if(is.fail()) 
    return IO_FAIL;
   
  istrstream str(m_currentline);
  string test;
  str >> test;
  if(test == m_nextsection)
    return IO_NEXTSECTION;
    
        
  return IO_GOOD;
}

void SSHeroReader::DoNextSection(ifstream &is)
{
  if(m_nextsection == c_VertexSection)
    {
      m_nextsection = c_FaceSection;
      ReadVertices(is);
    }
  else if(m_nextsection == c_FaceSection)
    {
      ReadFaces(is);
      m_nextsection = c_FaceSection;
    }
}

void SSHeroReader::ReadVertices(ifstream &is)
{
  ReadLine(is);

  c = 0;
  while(ReadLine(is) == IO_GOOD)
    {
      istrstream strtmp(m_currentline);
      string test;
      strtmp >> test;
        
      if(test == "]")
        break;

      GetVertices();       
      c++;
        
    }

  cout << "vertices: " << 3*c << endl;
}

void SSHeroReader::GetVertices()
{
  string curline(m_currentline), tmp;

  int lastp=0, newp=0;
  double x,y,z;
    
  do
    {
      newp = curline.find(",", lastp);
        
      if(newp == -1)
        return;
        
      tmp = curline.substr(lastp, newp - lastp);
      istrstream strtmp(tmp.c_str());
      strtmp >> x >> y >> z;
      m_vertices.push_back(SSVertex(x,y,z));
      lastp = newp + 1;
      if(c==0) {
        maxx = minx = x; maxy = miny = y; maxz = minz = z;
      } else {
        maxx = MAX(maxx,x); minx = MIN(minx,x);
        maxy = MAX(maxy,y); miny = MIN(miny,y);
        maxz = MAX(maxz,z); minz = MIN(minz,z);
      }
    }
  while(true);
}

void SSHeroReader::ReadFaces(ifstream &is)
{
  ReadLine(is);

  while(ReadLine(is) == IO_GOOD)
    {
      istrstream strtmp(m_currentline);
      string test;
      strtmp >> test;
        
      if(test == "]")
        break;

      GetFaceIndex();       
    }
}

void SSHeroReader::GetFaceIndex()
{
  string curline(m_currentline), tmp;
  int lastp=0, newp=0, x = 0;
  vector<int> facedata;
    
  do
    {     
      do
        {
          newp = curline.find(",", lastp);
            
          if(newp == -1)
            return;

          tmp = curline.substr(lastp, newp - lastp);

          if(tmp == "-1")
            {
              lastp = newp + 2;
              break;
            }

          istrstream strtmp(tmp.c_str());
          strtmp >> x;
          facedata.push_back(x);
          lastp = newp + 2;
        }
      while(true);
      m_faces.insert(make_pair(m_numberoffaces, facedata));
      facedata.clear();
      m_numberoffaces++;
    }
  while(true);
}

bool SSHeroReader::BuildModel(vector<SSLath*> &f)
{
  map<int, vector<int> >::const_iterator c_mapiter;
  vector<SSVertex> face_vertices;
  vector<int> facenumbers;
  vector<int>::const_iterator c_veciter;
  vector< vector<int> > hashtable( c_HashSize, vector<int>() );
  SSLath *facelath=0;

  int c=0, hashentry=0;
  double dmax, dx, dy, dz, scale, tx, ty, tz;

  c_mapiter = m_faces.begin();

  // Scale and translate model
  dx = maxx-minx;
  dy = maxy-miny;
  dz = maxz-minz;
  dmax = MAX(MAX(dx,dy),dz);
  scale = 3.0/dmax;
  tx = -(minx+dx/2.0);
  ty = -(miny+dy/2.0);
  tz = -(minz+dz/2.0);

  cout << "maxx = " << maxx << "   minx = " << minx
       << "   scale = " << scale << "   tx = " << tx << "   ty = " << ty 
       << "   tz = " << tz << endl;

  while(c_mapiter != m_faces.end())
    {
      facenumbers = c_mapiter->second;
      c_veciter = facenumbers.begin();
      while(c_veciter != facenumbers.end())
        {
          face_vertices.push_back((m_vertices[*c_veciter] + SSVertex(tx,ty,tz))*scale);
          c_veciter++;
        }

      facelath  = SSQModel::Instance()->BuildFace(face_vertices,f);
      for(int i = 0; i < facenumbers.size(); i++)
        {
          int h = hash(facenumbers[i], facenumbers[(i+1) % facenumbers.size()]);
          hashtable[h].push_back(c_mapiter->first);
        }
    
      face_vertices.clear();

      c_mapiter++;
      c++;
    }

  ConnectFaces(hashtable, f);
  cout << "Faces: " << c << endl;
  return true;
}

SSHeroReader::Match SSHeroReader::FindMatch(vector<vector<int> > &hashtable,
                                            vector<SSLath*> &faces,
                                            int v1,
                                            int v2)
{
  int h = hash(v1, v2);
  vector<int> faceindices = hashtable[h];
  Match m;

  for(vector<int>::const_iterator it=faceindices.begin(); it!=faceindices.end(); it++) {
    vector<int> vertexnums = m_faces[*it];
    for(int i=0; i<vertexnums.size(); i++)
      if(vertexnums[i] == v2 && vertexnums[(i+1)%vertexnums.size()] == v1) {
        // Match
        m.edge = (i+1)%vertexnums.size();
        m.face = faces[*it];
        return m;
      }
  }
  m.edge = 0;
  m.face = NULL;
  return m;
}

void SSHeroReader::ConnectFaces(vector< vector<int> > &hashtable, vector<SSLath*> &faces)
{
  map<int, vector<int> >::const_iterator c_mapiter;
  vector<int> vertexnumbers;
  SSLath *face3;
  c_mapiter = m_faces.begin();
  
  while(c_mapiter != m_faces.end())
    {
      vertexnumbers = c_mapiter->second;
      
      for(int i = 0; i < vertexnumbers.size(); i++)
        {
          SSLath *face = faces[c_mapiter->first];
          Match m = FindMatch(hashtable, faces, vertexnumbers[i], 
                              vertexnumbers[(i+1)%vertexnumbers.size()]);
          if(!m.face) {
            for(int j=0; j<i+1; j++) face = face->ccwf();
            face->setcwknotspc(0.0);
            face->cwf()->setccwknotspc(0.0);
            continue;
          } else {
            SSLath *f1 = face, *f2 = m.face;
            for(int j=0; j<i; j++) f1 = f1->ccwf();
            if(m.edge == 0) f2 = m.face->cwf();
            else
              for(int j=0; j<m.edge-1; j++) f2 = f2->ccwf();
            if(f1->ccwf()->ccwv()) {
              if(f1->ccwf()->ccwv() != f2) {
                f1->setccwknotspc(0.0);
                f1->ccwf()->setcwknotspc(0.0);
                f2->setccwknotspc(0.0);
                f2->ccwf()->setcwknotspc(0.0);
                f1->ccwf()->ccwv()->setccwknotspc(0.0);
                f1->ccwf()->ccwv()->ccwf()->setcwknotspc(0.0);
              }
              continue;
            }
            if(f2->ccwf()->ccwv()) {
              if(f2->ccwf()->ccwv() != f1) {
                f1->setccwknotspc(0.0);
                f1->ccwf()->setcwknotspc(0.0);
                f2->setccwknotspc(0.0);
                f2->ccwf()->setcwknotspc(0.0);
                f2->ccwf()->ccwv()->setccwknotspc(0.0);
                f2->ccwf()->ccwv()->ccwf()->setcwknotspc(0.0);
              }
              continue;
            }
            SSQModel::Instance()->ConnectFaces(face, (i+1)%vertexnumbers.size(), m.face, m.edge);

          }
        }
      c_mapiter++;
    }
}

void SSHeroReader::DumpAll()
{
  ofstream ofile;

  ofile.open("dump.txt");

  if(!ofile.is_open())
    {
      cout << "Dump file open error" << endl;
      return;
    }


  map<int, vector<int> >::const_iterator c_mapiter;
  vector<int> facenumbers;
  vector<int>::const_iterator c_veciter;
  vector<SSVertex>::const_iterator c_vertiter;

  c_mapiter = m_faces.begin();

  while(c_mapiter != m_faces.end())
    {

      ofile << "Face " << c_mapiter->first << ": ";

      facenumbers = c_mapiter->second;

      c_veciter = facenumbers.begin();
      while(c_veciter != facenumbers.end())
        {
          ofile << *c_veciter << " ";
          c_veciter++;
        }

      ofile << endl;

      c_mapiter++;
    }

  ofile << endl << "Vertices " << m_vertices.size() << endl;

  c_vertiter = m_vertices.begin();
  int c;

  while(c_vertiter != m_vertices.end())
    {
      c = 0;
      while(c <= 2 && c_vertiter != m_vertices.end())
        {
          ofile << *c_vertiter << " ";
          c++;
          c_vertiter++;
        }

      ofile << endl;
    }
}

int SSHeroReader::hash(int v1, int v2)
{
  return (v1 + v2) % c_HashSize;
}
