#if !defined _SSLath_HEADER 
#define _SSLath_HEADER

#include "SSVertex.h"

// Class representing a lath
class SSLath
{
public:
  // Next lath pointing to same vertex, clockwise and counter-clockwise, resp.
  virtual SSLath *cwv()=0;
  virtual SSLath *ccwv()=0;

  // Next lath in same face, clockwise and counter-clockwise resp.
  virtual SSLath *cwf()=0;
  virtual SSLath *ccwf()=0;

  // Associated lath on lower level
  virtual SSLath *nl()=0;

  // Vertex pointed to by this lath
  virtual SSVertex corner() const =0;

  // Output a string representation
  virtual ostream& put(ostream &s) const = 0;

  // Set next lath counter-clockwise around face or vertex, resp.
  virtual void ccwnextf(SSLath *lath)=0;
  virtual void ccwnextv(SSLath *lath)=0;

  // Set lath on lower level
  virtual void nextlevel(SSLath *lath)=0;

  // Check and mark lath as visited
  virtual bool isvisited() const { return false; };
  virtual void visited(bool b) {};

  // Update knot spacings, clockwise and counter-clockwise, resp.
  virtual void setcwknotspc(KnotSpacingType spc)=0;
  virtual void setccwknotspc(KnotSpacingType spc)=0;

  // Get knot spacings, clockwise and counter-clockwise, resp.
  virtual KnotSpacingType cwknotspc() const=0;
  virtual KnotSpacingType ccwknotspc() const=0;

  // Get knot spacing according to labelling defined by Sederberg et al.
  virtual KnotSpacingType knotspc(int s, int e, int r) const=0;

  // Destructor for clean-up
  virtual ~SSLath() {};
};

#endif
