#if defined(WIN32)
#pragma warning( disable : 4786 )
#endif

#include <cmath>
#include "SSLoop.h"
#include "SSQModel.h"

SSLoop::SSLoop()
{
}

SSLoop::~SSLoop()
{
}

SSLath *SSLoop::Subdivision(vector<SSLath*> &vec, SSLath *start)
{
  SSLoop::Subdivide(vec, start);
  return start->nl();
}

SSVector SSLoop::calcNormal(SSLath *face)
{
  SSLath *tmp = face, *tmp1, *tmp2, *last;
  SSVector v1, v2, norm;
  int i= 0, n = 0, countNonSmooth=0, isConical = 0, isCorner = 0, isCusp = 0;

  // Check for boundary.
  tmp1 = face;
  do {
    last = tmp1;
    tmp1 = tmp1->ccwv();
  } while(tmp1 && tmp1 != face);
  if(tmp1)
    last = NULL;

  tmp1 = face;
  do
    {
      n++;
      if(tmp1->cwknotspc() != 1.0) {
        if(tmp1->cwknotspc() == 0.0)
          countNonSmooth++;
        else if(tmp1->cwknotspc() == 2.0)
          isConical = 1;
        else if(tmp1->cwknotspc() == 3.0)
          isCusp = 1;
        else if(tmp1->cwknotspc() == 4.0)
          isCorner = 1;
      }
      tmp1 = tmp1->ccwv();
    }
  while(tmp1 && tmp1 != face);
  if(!tmp1) {
    tmp1 = face;
    do
      {
        n++;
        if(tmp1->ccwknotspc() != 1.0) {
          if(tmp1->ccwknotspc() == 0.0)
            countNonSmooth++;
          else if(tmp1->ccwknotspc() == 2.0)
            isConical = 1;
          else if(tmp1->ccwknotspc() == 3.0)
            isCusp = 1;
          else if(tmp1->ccwknotspc() == 4.0)
            isCorner = 1;
        }
        tmp1 = tmp1->cwv();
      }
    while(tmp1);
  }

  if(countNonSmooth > 2 || isCorner) {
    // CORNER CASE

    tmp1 = face;
    while(tmp1->cwv() != face && tmp1->ccwknotspc() != 0.0 && tmp1->ccwknotspc() != 4.0)
      tmp1 = tmp1->cwv();
    v1 = SSVector(tmp1->ccwf()->corner(), face->corner());
    tmp2 = tmp1;

    tmp1 = face;
    while(tmp1->ccwv() != face && tmp1->cwknotspc() != 0.0 && tmp1->cwknotspc() != 4.0)
      tmp1 = tmp1->ccwv();
    v2 = SSVector(tmp1->cwf()->corner(), face->corner());

    if(v1.len() < 1e-10 || v2.len() < 1e-10 || cross(v1,v2).len() < 1e-10) {
      v1 = SSVector(face->ccwf()->corner(), face->corner());
      v2 = SSVector(face->cwf()->corner(), face->corner());
    }
  } else if(countNonSmooth == 2) {
    // CREASE CASE

    SSVertex sv1, sv2;
    vector<SSVertex> v;
    double w;

    v.push_back(face->corner());
    tmp1 = face;
    while(tmp1 && tmp1->ccwknotspc() != 0.0)
      {
        tmp2 = tmp1;
        tmp1 = tmp1->cwv();
      } 
    if(!tmp1) {
      v.push_back(tmp2->ccwf()->corner());
      tmp1 = tmp2;
    } else {
      v.push_back(tmp1->ccwf()->corner());
    }
    do {
      v.push_back(tmp1->cwf()->corner());
      tmp1 = tmp1->ccwv();
    } while(tmp1 && tmp1->cwknotspc() != 0.0);
    if(tmp1) {
      v.push_back(tmp1->cwf()->corner());}
            
    v1 = SSVector(v[1],v[v.size()-1]);
    switch(v.size()) {
    case 3:
      v2 = SSVector(v[1],v[0]) + SSVector(v[2],v[0]);
      break;
    case 5:
      v2 = (SSVector(v[2]) + SSVector(v[3])) * 0.5
        - (SSVector(v[1]) + SSVector(v[0])*2.0 + SSVector(v[4])) * 0.25; 
      break;
    case 4:
      // Subdivision matrix defective - no set of spanning eigenvectors
      // Best solution: analyze with general eigenvectors.
      v2 = SSVector(v[1],v[0]) + SSVector(v[2],v[0]) + SSVector(v[3],v[0]);
      break;
    default:
      w = sin(c_PI/((double) (v.size()-2)))/(2.0*cos(c_PI/((double) (v.size()-2)))-2.0);
      v2 = w*(v[1]+v[v.size()-1]);
      for(i = 2; i<v.size()-1; i++)
        v2 += sin(c_PI*((double) (i-1))/(double)(v.size()-2)) * v[i];
      break;
    }
  } else if(isConical) {
    // CONICAL CASE

    // No tangent plane exists at central vertex
    // We use non-interpolated polygonal normal.
    v1 = SSVector(face->ccwf()->corner(), face->corner());
    v2 = SSVector(face->cwf()->corner(), face->corner());

  } else if(isCusp) {
    // CUSP CASE

    // Tangent plane collapses to tangent line.
    v1 = SSVector(face->corner());
    tmp1 = face;
    do {
      v1 -= SSVector(tmp1->ccwf()->corner()) * (1.0 / ((double) n));
      tmp1 = tmp1->ccwv();
    } while(tmp1 && tmp1 != face);
    v2 = SSVector(tmp1->ccwf()->corner(), tmp1->ccwf()->ccwf()->corner());

  } else {
    // SMOOTH OR DART CASE

    tmp1 = face;
    do {
      SSVector vi(tmp1->ccwf()->corner());
            
      v1 += vi * cos((2.0*c_PI*(double)i)/(double)n);
      v2 += vi * sin((2.0*c_PI*(double)i)/(double)n);
      i++;
      tmp1 = tmp1->cwv();
    } while(tmp1 && tmp1 != face);
    if(!tmp1) {
      tmp1 = face;
      do {
        SSVector vi(tmp1->cwf()->corner());
              
        v1 += vi * cos((2.0*c_PI*(double)i)/(double)n);
        v2 += vi * sin((2.0*c_PI*(double)i)/(double)n);
        i++;
        tmp1 = tmp1->ccwv();
      } while(tmp1 && tmp1 != face);
    }
    v1 = -v1;

  }
  norm = cross(v1, v2);
  norm.normalize();
  return norm;
}

void SSLoop::Subdivide(vector<SSLath*> &vec, SSLath *start)
{
  // If null or already subdivided just return
  if(start == 0 || start->isvisited())
    return;


  vector<SSVertex> vs, evs(3, SSVertex()), nvs;
  vector<SSLath *> els(4, (SSLath *) 0);
        
  SSLath *nf=0, *tmp1=0, *tmp2=0, *tmp3=0, *last=0;
  SSVertex ny, tmp;
        
  int i,k, k2, neighbourcountNonSmooth=0, neighbourIsCorner=0, countNonSmooth=0, 
    isConical=0, isCusp=0, isCorner=0, regularValency=6, neighbourRegularValency=6;
  CoefficientType cosinus,coeff;

  // Check for boundary. If found, move start to clockwise-most non-null lath
  tmp1 = start;
  do
    {
      tmp2 = tmp1;
      tmp1 = tmp1->cwv();
    }
  while(tmp1 && tmp1 != start);
  if(!tmp1) {
    do {
      last = start;
      start = start->ccwv();
    } while(start);
    start = tmp2;
    regularValency = 4;
  }

  // Collect the vertices around our start point
  // Count number of non-smooth edges incident
  // Check for conical point
  tmp1 = start;
  do
    {
      tmp1->visited(true);
      vs.push_back(tmp1->ccwf()->corner());

      if(tmp1->ccwknotspc() != 1.0) {
        if(tmp1->ccwknotspc() == 0.0)
          countNonSmooth++;
        else if(tmp1->ccwknotspc() == 2.0)
          isConical = 1;
        else if(tmp1->ccwknotspc() == 3.0)
          isCusp = 1;
        else if(tmp1->ccwknotspc() == 4.0)
          isCorner = 1;
      }
   
      tmp1 = tmp1->ccwv();
    }
  while(tmp1 && tmp1 != start);
  // If on the boundary remember last neighbour
  if(last) {
    vs.push_back(last->cwf()->corner());
          
    if(last->cwknotspc() != 1.0) {
      if(last->cwknotspc() == 0.0)
        countNonSmooth++;
      else if(last->cwknotspc() == 2.0)
        isConical = 1;
      else if(last->cwknotspc() == 3.0)
        isCusp = 1;
      else if(last->cwknotspc() == 4.0)
        isCorner = 1;
    }
  }
        
  k = vs.size();
  tmp = SSVertex();

  // Find appropriate vertex rule. Options are:
  // CREASE    - on a crease
  // CORNER    - endpoint of two or more creases
  // CONICAL   - conical point
  // CUSP      - cusp point
  // DART      - end point of one crease
  // SMOOTH    - smooth point

  if(countNonSmooth > 2 || isCorner) {
    // CORNER CASE

    isCorner = 1;
    nvs.push_back(start->corner());
  } else if(countNonSmooth == 2) {
    // CREASE CASE

    tmp1 = start;
    do {
      if(tmp1->ccwknotspc() == 0.0)
        tmp = tmp + 0.125 * tmp1->ccwf()->corner();
      tmp1 = tmp1->ccwv();
    } while(tmp1 && tmp1 != start);
    // If on boundary remember last edge
    if(last) {
      if(last->cwknotspc() == 0.0)
        tmp = tmp + 0.125 * last->cwf()->corner();
    }
    tmp = tmp + 0.75 * start->corner();
    nvs.push_back(tmp);
  } else if(isConical) {
    // CONICAL CASE

    coeff = ((2.0/(8.0*(CoefficientType) k)) * 
             (1.0 - cos(2.0*c_PI/(CoefficientType) k)));
          
    // Calculate the new vertex
    for(i=0; i<k; i++) 
      tmp = tmp + vs[i]*coeff;
        
    tmp = tmp + start->corner()*(1.0 - (CoefficientType)k*coeff);
    nvs.push_back(tmp);

  } else if(isCusp) {
    // CUSP CASE

    // New vertex identical to old
    nvs.push_back(start->corner());

  } else {
    // SMOOTH OR DART CASE (DART IF countNonSmooth==1)

    // Calculate new even vertex coefficient
    cosinus = 0.375 + 0.25 * cos((2.0*c_PI)/(CoefficientType)k);
    coeff = (CoefficientType) (1.0/(CoefficientType)k) * (0.625 - cosinus*cosinus);
          
    // Calculate the new vertex
    for(i=0; i<k; i++) 
      tmp = tmp + vs[i]*coeff;
        
    tmp = tmp + start->corner()*(1.0 - (CoefficientType)k*coeff);
    nvs.push_back(tmp);
  }       

  // Calculate the new edge vertices. Options:
  // SMOOTH    - smooth edge, possibly incident upon dart, conical or cusp vertex
  // ORDINARY CREASE - edge between two ordinary vertices, 
  //                   or between two extraordinary vertices
  // SPECIAL CREASE  - edge between ordinary and extraordinary vertices
        
  tmp1 = start;
  do
    {
      k2 = neighbourIsCorner = neighbourcountNonSmooth = 0;
      neighbourRegularValency = 6;
      tmp2 = tmp1->ccwf();
      do {
        // Neighbour is on the boundary if this is true
        if(!tmp2->ccwv())
          {
            neighbourRegularValency = 4;
            tmp3 = tmp1->ccwf();
            do
              {
                k2++;
                if(tmp3->ccwknotspc() == 0.0)
                  neighbourcountNonSmooth++;
                if(tmp3->ccwknotspc() == 4.0)
                  neighbourIsCorner = 1;
                tmp3 = tmp3->cwv();
              }
            while(tmp3);
          }

        k2++; // Valency of neighbour vertex

        if(tmp2->cwknotspc() == 0.0)
          neighbourcountNonSmooth++;
        if(tmp2->cwknotspc() == 4.0)
          neighbourIsCorner = 1;
        tmp2 = tmp2->ccwv();
      } while(tmp2 && tmp2 != tmp1->ccwf());
      if(neighbourcountNonSmooth > 2)
        neighbourIsCorner = 1;
            
      if((neighbourcountNonSmooth==1 || isConical || isCusp || countNonSmooth==1
          || tmp1->ccwknotspc()==1.0) && tmp1->ccwf()->ccwv()) {
        // SMOOTH CASE

        tmp = tmp1->ccwf()->corner()*0.375;
        tmp = tmp + tmp1->cwf()->corner()*0.125;
        tmp = tmp + tmp1->ccwf()->ccwv()->ccwf()->ccwf()->corner()*0.125;
        tmp = tmp + tmp1->corner()*0.375;
        nvs.push_back(tmp);
      } else if((tmp1->ccwknotspc() == 4.0 || tmp1->ccwknotspc() == 0.0) && 
                (isCorner || k!=regularValency) && k2==neighbourRegularValency && 
                !neighbourIsCorner) {
        // SPECIAL CREASE CASE 1 - central vertex extraordinary

        tmp = tmp1->corner() * 0.375 + tmp1->ccwf()->corner() * 0.625;
        nvs.push_back(tmp);
      } else if((tmp1->ccwknotspc() == 4.0 || tmp1->ccwknotspc() == 0.0) && !isCorner && 
                k==regularValency && (neighbourIsCorner || k2!=neighbourRegularValency)) {
        // SPECIAL CREASE CASE 2 - adjacent vertex extraordinary

        tmp = tmp1->corner() * 0.625 + tmp1->ccwf()->corner() * 0.375;
        nvs.push_back(tmp);
      } else {
        // ORDINARY CREASE

        tmp = tmp1->corner() * 0.5 + tmp1->ccwf()->corner() * 0.5;
        nvs.push_back(tmp);
      }
      tmp1 = tmp1->ccwv();
    }
  while(tmp1 && tmp1 != start);
  // If on boundary remember last edge
  if(last) {
    tmp1 = last;
    k2 = neighbourIsCorner = neighbourcountNonSmooth = 0;
    neighbourRegularValency = 4;
    tmp2 = tmp1->cwf();
    k2++;
    if(tmp2->ccwknotspc() == 0.0)
      neighbourcountNonSmooth++;
    if(tmp2->ccwknotspc() == 4.0)
      neighbourIsCorner = 1;
    do {
      k2++; // Valency of neighbour vertex
                
      if(tmp2->cwknotspc() == 0.0)
        neighbourcountNonSmooth++;
      if(tmp2->cwknotspc() == 4.0)
        neighbourIsCorner = 1;
      tmp2 = tmp2->ccwv();
    } while(tmp2);
    if(neighbourcountNonSmooth > 2)
      neighbourIsCorner = 1;
          
    if((tmp1->cwknotspc() == 4.0 || tmp1->cwknotspc() == 0.0) 
       && (isCorner || k!=regularValency) && k2==neighbourRegularValency 
       && !neighbourIsCorner) {
      // SPECIAL CREASE CASE 1 - central vertex extraordinary
            
      tmp = tmp1->corner() * 0.375 + tmp1->cwf()->corner() * 0.625;
      nvs.push_back(tmp);
    } else if((tmp1->cwknotspc() == 4.0 || tmp1->cwknotspc() == 0.0)
              && !isCorner && k==regularValency 
              && (neighbourIsCorner || k2!=neighbourRegularValency)) {
      // SPECIAL CREASE CASE 2 - adjacent vertex extraordinary
            
      tmp = tmp1->corner() * 0.625 + tmp1->cwf()->corner() * 0.375;
      nvs.push_back(tmp);
    } else {
      // ORDINARY CREASE
            
      tmp = tmp1->corner() * 0.5 + tmp1->cwf()->corner() * 0.5;
      nvs.push_back(tmp);
    }
  }

  // Build the new faces around the start vertex and connect to the next level
  tmp1 = start;
  evs[0] = nvs[0];
  i=1; k = vs.size();
  do
    {
      evs[1] = nvs[i];
      evs[2] = nvs[i%k+1];
      nf = SSQModel::Instance()->BuildFace(evs, vec);
      nf->setcwknotspc(tmp1->cwknotspc());
      nf->setccwknotspc(tmp1->ccwknotspc());
      if(tmp1->cwknotspc() == 0.0 || tmp1->cwknotspc() == 4.0)
        nf->cwf()->setccwknotspc(0.0);
      if(tmp1->ccwknotspc() == 0.0 || tmp1->ccwknotspc() == 4.0)
        nf->ccwf()->setcwknotspc(0.0);
      tmp1->nextlevel(nf->ccwf());
      tmp1 = tmp1->ccwv();
      i++;
    }
  while(tmp1 && tmp1 != start);
        
  // Connect the new faces
  tmp1 = start;
  do
    {
      if(tmp1->ccwv())
        SSQModel::Instance()->ConnectFaces(tmp1->nl(), 2, tmp1->ccwv()->nl(), 0);
      tmp1 = tmp1->ccwv();
    }
  while(tmp1 && tmp1 != start);

          
  // Start subdividing
  tmp1 = start;
  do 
    {
      SSLoop::Subdivide(vec, tmp1->ccwf());
      tmp1 = tmp1->ccwv();
    } 
  while(tmp1 && start != tmp1);
  if(last)
    SSLoop::Subdivide(vec, last->cwf());

  // Build and Connect new face in the middle
  tmp1 = start;
  do
    {
      if(!tmp1->nl()->ccwf()->ccwv() && tmp1->ccwf()->nl() && tmp1->ccwf()->ccwf()->nl())
        {
          evs[0] = tmp1->nl()->corner();
          evs[1] = tmp1->ccwf()->nl()->corner();
          evs[2] = tmp1->ccwf()->ccwf()->nl()->corner();
          nf = SSQModel::Instance()->BuildFace(evs, vec);
                
          SSQModel::Instance()->ConnectFaces(tmp1->nl(), 1, nf, 0);
          SSQModel::Instance()->ConnectFaces(tmp1->ccwf()->nl(), 1, nf, 1);
          SSQModel::Instance()->ConnectFaces(tmp1->ccwf()->ccwf()->nl(), 1, nf, 2);
        }

      tmp1 = tmp1->ccwv();
    }
  while(tmp1 && tmp1 != start);
                
}
