#if !defined _SSModel_HEADER 
#define _SSModel_HEADER

#include "SSMaster.h"
#include "SSLath.h"
#include "SSVertex.h"
#include "ErrorDescription.h"

// Base class for representing 3D polygonal models
class SSModel
{
public:

  // Convenience methods for building a face with or without knot spacings
  // The new laths are appended to vec
  virtual SSLath *BuildFace(const vector<SSVertex> vertices, vector<SSLath*> &vec)=0;
  virtual SSLath *BuildFace(const vector<SSVertex> vertices,
                            const vector<KnotSpacingType> knots,
                            vector<SSLath*> &vec)=0;

  // Convenience method for stitching together two faces across numbered edges
  // Edges are numbered counterclockwise starting with 0 on the clockwise edge 
  // from the lath pointer
  virtual ErrorDescription ConnectFaces(SSLath* f1, int edgeno1, 
                                        SSLath* f2, int edgeno2)=0;

  // Collects the vertices and optionally knot spacings in the face given by start
  virtual ErrorDescription FaceVertices(vector<SSVertex> &vertices, SSLath* start)=0;
  virtual ErrorDescription FaceVertices(vector<SSVertex> &vertices, 
                                        vector<KnotSpacingType> &knots, 
                                        SSLath* start)=0;

  // Loads a model from the file filename
  virtual ErrorDescription Init(const string &filename)=0;

  // Prepares the model for a new subdivision scheme
  virtual ErrorDescription SwitchMethod(MethodType newmethod)=0;

  // Subdivides the model using the specified scheme
  virtual ErrorDescription Subdivide(MethodType m, int levels)=0;

  // Draws a specific subdivision level of the model using OpenGL
  // Flags control the rendering
  virtual ErrorDescription Draw(int level, int flags)=0;

  // Return the number of faces in the given level of the model
  virtual long NumberOfFaces(int level) { return 0; };

  // Delete a level of the model
  virtual void DeleteLevel(LevelType l)=0;

  // Destructor for clean-up
  virtual ~SSModel() {};

protected:
  SSModel() {};

};

#endif
