#if !defined _SSQModel_HEADER 
#define _SSQModel_HEADER

#include "SSMaster.h"
#include "SSSubdivideMethods.h"
#include "SSModel.h"
#include "SSVector.h"
#include <GL/glut.h>

typedef vector<SSLath*> LathVector;
typedef pair<SSLath*, LathVector> Level;

// Specialization of SSModel
// See SSModel.h for method descriptions
class SSQModel : public SSModel
{
public:
  virtual SSLath *BuildFace(const vector<SSVertex> vertices,
                            vector<SSLath*> &vec);
  virtual SSLath *BuildFace(const vector<SSVertex> vertices,
                            const vector<KnotSpacingType> knots,
                            vector<SSLath*> &vec);

  virtual ErrorDescription ConnectFaces(SSLath* f1, int edgeno1, 
                                        SSLath* f2, int edgeno2);

  virtual ErrorDescription FaceVertices(vector<SSVertex> &vertices, SSLath* start);
  virtual ErrorDescription FaceVertices(vector<SSVertex> &vertices, 
                                        vector<KnotSpacingType> &knots,
                                        SSLath* start);

  virtual ErrorDescription Init(const string &filename);

  virtual SSVector calcNormal(SSLath *face);

  virtual ErrorDescription Draw(int level, int flags);
        
  virtual ErrorDescription SwitchMethod(MethodType newmethod);

  virtual ErrorDescription Subdivide(MethodType m, int levels);

  virtual void DeleteLevel(LevelType l);

  virtual long NumberOfFaces(int level) 
  { 
    if(level < 0 || level > m_curlevel)
      level = 0;

    return m_Levels[level].second.size(); 
  };

  static SSQModel *Instance();
        
  virtual ~SSQModel();

  // Triangulates the current model by inserting new vertices at the
  // centroid of all faces
  virtual void Triangulate();
  
protected:
  SSQModel();

private:
  void DrawFace(SSLath *start, int flags);

  static auto_ptr<SSQModel> m_instance;
  SSSubdivideMethods *m_method;
  LevelType m_curlevel;
  map<LevelType, Level> m_Levels;
  MethodType m_curmethod;
  string m_filename;
};

#endif
