#if !defined _SSSubdivideMethods_HEADER 
#define _SSSubdivideMethods_HEADER

#include "ErrorDescription.h"
#include "SSCornerLath.h"
#include "SSVector.h"

// Base class for implementing subdivision schemes
class SSSubdivideMethods
{

public:
  virtual SSLath* Subdivision(vector<SSLath*> &vec, SSLath *start)=0;

  // Generic interpolated polygonal normal
  // Overridden by classes providing scheme specific normals
  virtual SSVector calcNormal(SSLath *face)
  {
    SSVector norm(0.0, 0.0, 0.0);
    SSLath *tmp;
    tmp = face;
    int count = 0;
    do {
      SSVector v1(tmp->corner(), tmp->ccwf()->corner());
      SSVector v2(tmp->corner(), tmp->cwf()->corner());
      v1.normalize(); v2.normalize(); // Normalize to reduce numerical instability
      SSVector c = cross(v1, v2);
      norm += c;
      tmp = tmp->ccwv();
      count++;
      if(count > 200)
        exit(0);   // Valency > 200 indicates a malformed mesh and infinite loop
    } while(tmp && tmp != face);
    norm.normalize();
    return norm;
  };

  virtual ~SSSubdivideMethods() {};
};
        
#endif
