#if !defined _SSVertex_HEADER 
#define _SSVertex_HEADER

#ifndef _SSMaster_HEADER
#include "SSMaster.h"
#endif

class SSVertex
{

public:
  SSVertex();

  SSVertex(CoordType x,
           CoordType y,
           CoordType z);

  virtual ~SSVertex();

  CoordType x() const { return m_x;};
  CoordType y() const { return m_y;};
  CoordType z() const { return m_z;};

  bool operator==(const SSVertex& rhs)
  {
    return (m_x == rhs.x() && m_y == rhs.y() && m_z == rhs.z());
  }

  bool operator!=(const SSVertex& rhs)
  {
    return (m_x != rhs.x() || m_y != rhs.y() || m_z != rhs.z());
  }

  SSVertex& operator= (const SSVertex& rhs)
  {
    if(*this == rhs)
      return *this;

    m_x = rhs.x();
    m_y = rhs.y();
    m_z = rhs.z();

    return *this; 
  };

  SSVertex operator+ (const SSVertex& rhs)
  {
    return SSVertex(m_x + rhs.m_x, m_y + rhs.m_y, m_z + rhs.m_z);
  };

  void operator+= (const SSVertex& rhs)
  {
    m_x += rhs.m_x;
    m_y += rhs.m_y;
    m_z += rhs.m_z;
  }

  SSVertex operator* (const CoefficientType& c)
  {
    return  SSVertex(m_x * c, m_y * c, m_z * c); 
  }

  SSVertex operator/ (const CoefficientType& c)
  {
    if(c == 0) {
      cout << "Division by zero" << endl;
      throw "Div by zero";
    }

    return  SSVertex(m_x / c, m_y / c, m_z / c); 
  }

  void operator/= (const CoefficientType& c)
  {
    m_x /= c;
    m_y /= c;
    m_z /= c;
  }    
  
private:
  CoordType m_x, m_y, m_z;

  friend SSVertex operator*(const CoefficientType& c, const SSVertex& v);
};

inline SSVertex operator*(const CoefficientType& c, const SSVertex& v)
{
  return SSVertex(c*v.m_x, c*v.m_y, c*v.m_z);
}

ostream& operator<<(ostream&, const SSVertex & v);
ostream& operator<<(ostream&, const vector<SSVertex> & vec);

#endif
