#if defined(WIN32)
#pragma warning( disable : 4786 )
#endif

#ifndef _SSMaster_HEADER
#include "SSMaster.h"
#endif

#include <GL/glut.h>
#include "SSQModel.h"
#include "SSVector.h"
#include "SSText.h"
#include <iostream>
#if !defined(WIN32)
#include <unistd.h>
#include <cstdio>
#endif
#include <cmath>

#define VIEWER_VERSION "0.3"

static GLint ay = 0, ar = 0;
static GLfloat tz = -10, tx = 0.0, ty = 0.0;
int flags = SS_DRAW_OUTLINE;
bool hiddenlines = true, 
  depth = true, 
  lighting = true, 
  outline = true, 
  perspectiveproj=true,
  trace = false,
  FullScreen = false,
  exactnormal = false;

int maxlevel=4;
int curlevel=0;
int method=c_DooSabin;
int winWidth = 800, winHeight = 600;

enum MENU_ENUM
{
  M_DOOSABIN = 0,
  M_CATMULLCLARK,
  M_LOOP,
  M_QUADNURSS,
  M_CUBICNURSS,
  M_CULL,
  M_LIGHTING,
  M_PERSPECTIVE,
  M_OUTLINE,
  M_DEPTH,
  M_TRACE,
  M_FULL_SCREEN,
  M_QUIT
};

void draw_axis(SSVector axlen) 
{
  SSVector zero;
  SSVector tip1, tip2, tip3;
  double colorR[] = {1.0, 0.0, 0.0, 1.0};
  double colorG[] = {0.0, 1.0, 0.0, 1.0};
  double colorB[] = {0.0, 0.0, 1.0, 1.0};
                                    
  glPushAttrib(GL_ALL_ATTRIB_BITS);

  glDisable(GL_LIGHTING);
  glDisable(GL_COLOR_MATERIAL);
  glLineWidth(2.0f);

  glBegin(GL_LINES);
  glColor3ub(255,  0,  0);
  tip1 = SSVector(axlen[0], 0.0f, 0.0f);
  glVertex3f(zero[0], zero[1], zero[2]);
  glVertex3f(tip1[0], tip1[1], tip1[2]);
  
  glColor3ub(0, 255,  0);
  tip2 = SSVector(0.0f, axlen[1], 0.0f);
  glVertex3f(zero[0], zero[1], zero[2]);
  glVertex3f(tip2[0], tip2[1], tip2[2]);

  glColor3ub(0,  0, 255);
  tip3 = SSVector(0.0f, 0.0f, axlen[2]);
  glVertex3f(zero[0], zero[1], zero[2]);
  glVertex3f(tip3[0], tip3[1], tip3[2]);
  glEnd();

  glColor3ub(0,0,0);
  SSGLText *textX = new SSGLText(tip1, SSVector(.002f, .002f, .002f), " x");
  SSGLText *textY = new SSGLText(tip2, SSVector(.002f, .002f, .002f), " y");
  SSGLText *textZ = new SSGLText(tip3, SSVector(.002f, .002f, .002f), " z");
  
  textX->draw();
  textY->draw();
  textZ->draw();
  
  delete textX;
  delete textY;
  delete textZ;
  glPopAttrib();
}

void display_text(GLuint x, GLuint y, double scale, char* text) 
{
  double font_scale = 119.05f + 33.33f;

  glPushMatrix();
  glMatrixMode(GL_PROJECTION);
  glPushMatrix();
  glLoadIdentity();

  gluOrtho2D(0, winWidth, 0, winHeight);

  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();

  glTranslatef(x, y, 0);
  glScalef(scale / font_scale, scale / font_scale, 1);

  glPushAttrib(GL_CURRENT_BIT);
  glDisable(GL_LIGHTING);
  glColor3f(1, 1, 1);
  while (*text) glutStrokeCharacter(GLUT_STROKE_ROMAN, *text++);
 
  glPopMatrix();
 
  glMatrixMode(GL_PROJECTION);
  glPopMatrix();
 
  glMatrixMode(GL_MODELVIEW);
  glPopAttrib();
  if(lighting)
    glEnable(GL_LIGHTING);

}

void set_projection(void) 
{
  double aspect = (double) winWidth / (double) winHeight;
        
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  if (perspectiveproj) 
    {
      gluPerspective(35, aspect, 0.1, 700.0);
    }
  else 
    {
      glOrtho(-3*aspect, 3*aspect, -3, 3, 0.1, 20.0);
    }

  glMatrixMode(GL_MODELVIEW);
}

void init(const string &filename)
{
  GLfloat light_diffuse[]  = {0.6f, 0.6f, 1.0f, 1.0f};  
  GLfloat light_position[] = {0.0, 0.0, 1000.0, 1};

  float ambient[] = {0.5, 0.0, 0.0, 1.0};
  float diffuse[] = {1.0, 0.1, 0.1, 1.0};
  float specular[] = {1.0,1.0, 1.0, 1.0};
  float shine[] = {100.0};
  float emission[] = {0, 0, 0, 0};

  glMaterialfv(GL_FRONT, GL_EMISSION, emission);
  glMaterialfv(GL_FRONT, GL_AMBIENT, ambient);
  glMaterialfv(GL_FRONT, GL_DIFFUSE, diffuse);
  glMaterialfv(GL_FRONT, GL_SHININESS, shine);
  glMaterialfv(GL_FRONT, GL_SPECULAR, specular);
        
  glShadeModel(GL_SMOOTH);
  glEnable(GL_LIGHTING);
  glLightfv(GL_LIGHT0, GL_DIFFUSE, light_diffuse);
  glLightfv(GL_LIGHT0, GL_POSITION, light_position);
  
  glEnable(GL_LIGHT0);    
  glClearColor(.3, .3, .3, 1.0F);

  glFrontFace(GL_CCW);

  cout << "Reading model" << endl;
  SSQModel::Instance()->Init(filename);
  cout << "Model read" << endl;

        
  SSQModel::Instance()->Subdivide(method, maxlevel);

  cout << "Model subdivided" << endl;
}

void display(void)
{
  char text[128];

  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  glTranslatef(tx, ty, tz);
        
  glRotatef((GLfloat) ay, 0.0, 1.0, 0.0);
  glRotatef((GLfloat) ar, (GLfloat) cos(-ay*c_PI/180.0), 0.0, (GLfloat) -sin(-ay*c_PI/180.0));
        
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  if(hiddenlines) 
    glEnable(GL_CULL_FACE);
        
  if(depth)
    glEnable(GL_DEPTH_TEST);

  if (outline)
    glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
        
  if(lighting)
    {
      glEnable(GL_LIGHTING);
      glEnable(GL_LIGHT0);
    }

  SSVector axis = SSVector(0.5, 0.5, 0.5);        
  draw_axis(axis);

  if(method == c_DooSabin)
    sprintf(text, "%s", "DooSabin");
  else if (method == c_CatmullClark)
    sprintf(text, "%s", "CatmullClark");
  else if (method == c_Loop)
    sprintf(text, "%s", "Loop");
  else if (method == c_QuadNURSS)
    sprintf(text, "%s", "Quad Nurss");
  else if (method == c_CubicNURSS)
    sprintf(text, "%s", "Cubic Nurss");
  else
    sprintf(text, "%s", "Unknown");

  display_text(10, 30, 20, text);
        
  SSQModel::Instance()->SwitchMethod(method);
        
  SSQModel::Instance()->Draw(curlevel, flags);

  long faces = SSQModel::Instance()->NumberOfFaces(curlevel);
        
  sprintf(text, "%d", faces);
        
  display_text(10, 10, 20, text);

  if(trace) 
    {
      trace = false;
      flags ^= SS_DRAW_TRACE;
    }
        
  glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_FALSE);
  glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
  glDisable(GL_CULL_FACE);
  glDisable(GL_LIGHT0);
  glDisable(GL_DEPTH_TEST);

  glutSwapBuffers();
}

void reshape(int width, int height)
{
  glViewport(0, 0, width, height);
  winWidth = width;
  winHeight = height;
  set_projection();
}

void keyboard(unsigned char key, int x, int y)
{
  switch(key) {
  case 27:
    exit(0);
    break;
  case 'h':
    hiddenlines = !hiddenlines;
    break;
  case 'd':
    depth = !depth;
    break;
  case 'o':
    outline = !outline;
    flags ^= SS_DRAW_OUTLINE;
    break;
  case 'p':
    perspectiveproj = !perspectiveproj;
    set_projection();
    break;
  case 'n':
    exactnormal = !exactnormal;
    flags ^= SS_DRAW_EXACT_NORMALS;
    break;
          
  case 'l':
    lighting = !lighting;
    break;
  }
  glutPostRedisplay();
}

void specKeyboard(int key, int x, int y)
{
  int s = glutGetModifiers();
  switch(key) {
  case GLUT_KEY_LEFT:
    if(s == GLUT_ACTIVE_CTRL)
      tx -= 0.2;
    else
      ay = (ay + 5) % 360;
    break;
  case GLUT_KEY_RIGHT:
    if(s == GLUT_ACTIVE_CTRL)
      tx += 0.2;
    else
      ay = (ay - 5) % 360;
    break;
  case GLUT_KEY_UP:
    if(s == GLUT_ACTIVE_CTRL)
      ty += 0.2;
    else
      ar = (ar + 5) % 360;
    break;
  case GLUT_KEY_DOWN:
    if(s == GLUT_ACTIVE_CTRL)
      ty -= 0.2;
    else
      ar = (ar - 5) % 360;
    break;
  case GLUT_KEY_PAGE_UP:
    tz -= 0.2;
    break;
  case GLUT_KEY_PAGE_DOWN:
    tz += 0.2;
    break;
  case GLUT_KEY_HOME:
  case GLUT_KEY_F12:
    if(curlevel < maxlevel)
      curlevel++;
    else
      curlevel = 0;
    glutPostRedisplay();
    break;
  case GLUT_KEY_END:
  case GLUT_KEY_F11:
    if(curlevel > 0)
      curlevel--;
    else
      curlevel = 0;
    glutPostRedisplay();
    break;
  }
  glutPostRedisplay();
}

void menu_fun(int value) 
{
  switch (value) 
    {
    case M_LOOP:
      method = c_Loop;
      break;
    case M_CATMULLCLARK:
      method = c_CatmullClark;
      break;
    case M_DOOSABIN:
      method = c_DooSabin;
      break;
    case M_QUADNURSS:
      method = c_QuadNURSS;
      break;
    case M_CUBICNURSS:
      method = c_CubicNURSS;
      break;
  
    case M_CULL:
      hiddenlines = !hiddenlines;     
      break;  
    case M_LIGHTING: 
      lighting = !lighting;   
      break;
                        
    case M_OUTLINE: 
      outline = !outline;
      flags ^= SS_DRAW_OUTLINE;
      break;
                        
    case M_PERSPECTIVE: 
      perspectiveproj = !perspectiveproj;     
      set_projection();
      break;
                        
    case M_DEPTH: 
      depth = !depth; 
      break;
    case M_TRACE:
      trace = !trace;
      flags ^= SS_DRAW_TRACE;
      break;
                        
    case M_FULL_SCREEN:
      FullScreen = !FullScreen;

      if (FullScreen)
        glutFullScreen();
      else
        glutReshapeWindow(800, 600);
      break;
                        
    case M_QUIT:
      exit(0);
      break;
    }

  glutPostRedisplay();
}

void init_menu() 
{
  glutCreateMenu(menu_fun);
  glutAddMenuEntry("Doo Sabin", M_DOOSABIN);
  glutAddMenuEntry("Catmull Clark", M_CATMULLCLARK);
  glutAddMenuEntry("Loop", M_LOOP);
  glutAddMenuEntry("Quad Nurss", M_QUADNURSS);
  glutAddMenuEntry("Cubic Nurss", M_CUBICNURSS);
  glutAddMenuEntry("", -1);
  glutAddMenuEntry("", -1);
  glutAddMenuEntry("Toggle Face Culling", M_CULL);
  glutAddMenuEntry("Toggle Lighting", M_LIGHTING);
  glutAddMenuEntry("Toggle Outlines", M_OUTLINE);
  glutAddMenuEntry("Toggle Perspective Proj", M_PERSPECTIVE);
  glutAddMenuEntry("Toggle Depth Test", M_DEPTH);
  glutAddMenuEntry("", -1);
  glutAddMenuEntry("Trace", M_TRACE);
  glutAddMenuEntry("Full screen", M_FULL_SCREEN);
  glutAddMenuEntry("", -1);
  glutAddMenuEntry("Exit", M_QUIT);
  glutAttachMenu(GLUT_RIGHT_BUTTON);
}

int main(int argc, char **argv)
{
  glutInit(&argc, argv);
  glutInitDisplayMode(GLUT_DOUBLE | GLUT_DEPTH );
  glutInitWindowSize(winWidth,winHeight);
  glutCreateWindow("Viewer " VIEWER_VERSION);
  if(argc > 2)
    maxlevel = strtol(argv[2], 0, 10);
  if(argc > 1)
    init(argv[1]);
  else {
    cerr << "No input file specified.\n" << endl;
    exit(1);
  }
  glutDisplayFunc(display);
  glutReshapeFunc(reshape);
  glutKeyboardFunc(keyboard);
  glutSpecialFunc(specKeyboard);

  init_menu();

  glutMainLoop();
  return 0;
}
